from stftpitchshift import StftPitchShift
from stftpitchshift import __version__ as version
from stftpitchshift.io import read, write
from stftpitchshift.stft import stft, istft, spectrogram

import click
import matplotlib.pyplot as plot
import numpy as np


@click.command('stftpitchshift', help='STFT based multi pitch shifting with optional formant preservation', no_args_is_help=True, context_settings=dict(help_option_names=['-h', '--help']))
@click.version_option(version=version, message='%(version)s')
@click.option('-i', '--input', required=True, help='input .wav file name')
@click.option('-o', '--output', required=True, help='output .wav file name')
@click.option('-p', '--pitch', default='1.0', show_default=True, help='fractional pitch shifting factors separated by comma')
@click.option('-q', '--quefrency', default='0.0', show_default=True, help='optional formant lifter quefrency in milliseconds')
@click.option('-w', '--window', default=1024, show_default=True, help='sfft window size')
@click.option('-v', '--overlap', default=32, show_default=True, help='stft window overlap')
@click.option('-d', '--debug', is_flag=True, default=False, help='plot spectrograms before and after processing')
def main(input, output, pitch, quefrency, window, overlap, debug):

    x, samplerate = read(input)

    factors = list(set(float(factor) for factor in pitch.split(',')))
    quefrency = float(quefrency) * 1e-3

    framesize = window
    hopsize = window // overlap

    pitchshifter = StftPitchShift(framesize, hopsize, samplerate)

    channels = x.shape[-1] if x.ndim > 1 else 1

    x = x[:, None] if channels == 1 else x

    y = np.stack([
        pitchshifter.shiftpitch(x[:, channel], factors, quefrency)
        for channel in range(channels)
    ], axis=-1)

    write(output, y, samplerate)

    if debug:

        for channel in range(channels):

            framesX = stft(x[:, channel], framesize, hopsize)
            framesY = stft(y[:, channel], framesize, hopsize)

            figure = plot.figure(f'Channel {channel+1}/{channels}')

            spectrogramX = figure.add_subplot(2, 1, 1, title='Input Spectrogram')
            spectrogram(framesX, framesize, hopsize, samplerate)

            spectrogramY = figure.add_subplot(2, 1, 2, title='Output Spectrogram')
            spectrogram(framesY, framesize, hopsize, samplerate)

            spectrogramX.get_shared_x_axes().join(spectrogramX, spectrogramY)
            spectrogramX.get_shared_y_axes().join(spectrogramX, spectrogramY)

            plot.tight_layout()

        plot.show()


if __name__ == '__main__':

    main()



import numpy as np


def linear(x, factor):

    if factor == 1:
        return x.copy()

    N = len(x)
    M = round(N * factor)

    y = np.zeros(x.shape, dtype=x.dtype)

    m = np.arange(min(M, N))
    n = m / factor

    n0 = np.floor(n).astype(int)
    n1 = np.ceil(n).astype(int)

    mask = (0 <= n0) & (n0 < N) & (0 <= n1) & (n1 < N)

    m = m[mask]
    n = n[mask]
    n0 = n0[mask]
    n1 = n1[mask]

    y0 = x[n0]
    y1 = x[n1]

    a = n - n0
    b = n1 - n0

    mask = (n0 == n1)

    a[mask] = 1
    b[mask] = 2

    i = a / b

    y[m] = y0 * (1 - i) + y1 * i

    return y


def bilinear(x0, x1, factor):

    y0 = linear(x0, factor)
    y1 = linear(x1, factor)

    return (y0 + y1) / 2



from stftpitchshift.cepster import lifter
from stftpitchshift.pitcher import shiftpitch
from stftpitchshift.stft import stft, istft, spectrogram
from stftpitchshift.vocoder import encode, decode

import numpy as np


class StftPitchShift:
    '''
    Short-time Fourier transform (STFT) based pitch shifting.
    '''

    def __init__(self, framesize, hopsize, samplerate):
        '''
        :param framesize: The STFT frame size in samples.
        :param hopsize: The STFT hop size in samples.
        :param samplerate: The sample rate of the signal in hertz.
        '''

        self.framesize = framesize
        self.hopsize = hopsize
        self.samplerate = samplerate

    def shiftpitch(self, input, factors = 1, quefrency = 0):
        '''
        :param input: The input signal.
        :param factors: The fractional pitch shifting factors.
        :param quefrency: The optional formant lifter quefrency in seconds.
        :return: The output signal of the equal size.
        '''

        def isnotnormal(x):
            return (np.isinf(x)) | \
                   (np.isnan(x)) | \
                   (abs(x) < np.finfo(x.dtype).tiny)

        framesize = self.framesize
        hopsize = self.hopsize
        samplerate = self.samplerate

        factors = np.asarray(factors).flatten()
        quefrency = int(quefrency * samplerate)

        frames = stft(input, framesize, hopsize)

        if quefrency:

            frames = encode(frames, framesize, hopsize, samplerate)

            envelopes = lifter(frames, quefrency)

            mask = isnotnormal(envelopes)

            frames.real /= envelopes
            frames.real[mask] = 0

            frames = shiftpitch(frames, factors)

            frames.real *= envelopes
            frames.real[mask] = 0

            frames = decode(frames, framesize, hopsize, samplerate)

        else:

            frames = encode(frames, framesize, hopsize, samplerate)
            frames = shiftpitch(frames, factors)
            frames = decode(frames, framesize, hopsize, samplerate)

        output = istft(frames, framesize, hopsize)

        output.resize(np.shape(input))

        return output



from numpy.lib.stride_tricks import sliding_window_view

import matplotlib
import matplotlib.pyplot as plot
import numpy as np


def stft(x, framesize, hopsize):

    frames = sliding_window_view(x, framesize, writeable=False)[::hopsize]

    M, N = frames.shape

    data = np.zeros((M, N//2+1), complex)

    w = 0.5 - 0.5 * np.cos(2 * np.pi * np.arange(framesize) / framesize)

    for i, frame in enumerate(frames):

        data[i] = np.fft.rfft(w * frame, norm='forward')

        # keep dc and nyquist as is
        # data[i][0] = data[i][-1] = 0

    return data


def istft(frames, framesize, hopsize):

    M, N = frames.shape

    y = np.zeros((M * hopsize + framesize), float)

    data = sliding_window_view(y, framesize, writeable=True)[::hopsize]

    w = 0.5 - 0.5 * np.cos(2 * np.pi * np.arange(framesize) / framesize)
    w *= hopsize / np.sum(w**2)

    for i, frame in enumerate(frames):

        # zero dc and nyquist
        frame[0] = frame[-1] = 0

        data[i] += w * np.fft.irfft(frame, norm='forward')

    return y


def spectrogram(frames, framesize, hopsize, samplerate, xlim=None, ylim=None, clim=-120, cmap='inferno'):

    def lim():

        if xlim is not None:
            if isinstance(xlim, (list, tuple)):
                plot.xlim(xlim)
            else:
                plot.xlim(0, xlim)

        if ylim is not None:
            if isinstance(ylim, (list, tuple)):
                plot.ylim(ylim)
            else:
                plot.ylim(0, ylim)

        if clim is not None:
            if isinstance(clim, (list, tuple)):
                plot.clim(clim)
            else:
                plot.clim(clim, 0)

    with np.errstate(divide='ignore', invalid='ignore'):
        data = 20 * np.log10(np.abs(frames))

    timestamps = np.arange(frames.shape[0]) * hopsize / samplerate
    frequencies = np.arange(frames.shape[1]) * samplerate / framesize

    roi = (np.min(timestamps), np.max(timestamps), np.min(frequencies), np.max(frequencies))

    colormap = matplotlib.cm.get_cmap(cmap)
    colormap.set_bad(colormap(0))

    plot.imshow(data.T, extent=roi, cmap=colormap, aspect='auto', interpolation='nearest', origin='lower')

    plot.xlabel('s')
    plot.ylabel('Hz')

    colorbar = plot.colorbar()
    colorbar.set_label('dB')

    lim()

    return plot


def phasogram(frames, framesize, hopsize, samplerate, xlim=None, ylim=None, clim=None, cmap='twilight'):

    def lim():

        if xlim is not None:
            if isinstance(xlim, (list, tuple)):
                plot.xlim(xlim)
            else:
                plot.xlim(0, xlim)

        if ylim is not None:
            if isinstance(ylim, (list, tuple)):
                plot.ylim(ylim)
            else:
                plot.ylim(0, ylim)

        if clim is not None:
            if isinstance(clim, (list, tuple)):
                plot.clim(clim)
            else:
                plot.clim(-clim, +clim)
        else:
            plot.clim(-np.pi, +np.pi)

    data = np.angle(frames)

    timestamps = np.arange(frames.shape[0]) * hopsize / samplerate
    frequencies = np.arange(frames.shape[1]) * samplerate / framesize

    roi = (np.min(timestamps), np.max(timestamps), np.min(frequencies), np.max(frequencies))

    plot.imshow(data.T, extent=roi, cmap=cmap, aspect='auto', interpolation='nearest', origin='lower')

    plot.xlabel('s')
    plot.ylabel('Hz')

    colorbar = plot.colorbar()
    colorbar.set_label('rad')

    lim()

    return plot



import numpy
import os
import subprocess
import sys
import wave


def play(path, cmd='play', shell=False, wait=True):
    """
    Executes custom command with specified .wav file path as an argument.

    Parameters
    ----------
    path : string
        File path with or without the .wav extension.
    cmd : string
        Command to execute, e.g. play (http://sox.sourceforge.net).
    shell : bool
        If true, the command will be executed through the shell.
    wait : bool
        If true, wait for child process to terminate.
    """

    if path.startswith('~'):
        path = os.path.expanduser(path)

    if not path.lower().endswith('.wav'):
        path += '.wav'

    process = subprocess.Popen([cmd, path], shell=shell)

    if wait:
        return process.wait()
    else:
        return process.pid


def read(path):
    """
    Reads a .wav file.

    Parameters
    ----------
    path : string
        File path with or without the .wav extension.

    Returns
    -------
    data : ndarray
        Content of the .wav file.
    sr : integer
        Sample rate in hertz.
    """

    if path.startswith('~'):
        path = os.path.expanduser(path)

    if not path.lower().endswith('.wav'):
        path += '.wav'

    with wave.open(path, 'rb') as file:
        sr = file.getframerate()
        bytes = file.getsampwidth()
        channels = file.getnchannels()
        data = file.readframes(file.getnframes())

    assert bytes in [1, 2, 3, 4]
    bits = bytes * 8
    scaler = 2 ** (bits - 1) - 1

    data = numpy.frombuffer(data, dtype=numpy.uint8).reshape(-1, bytes)
    data = numpy.asarray([
        int.from_bytes(frame, signed=(bits != 8), byteorder=sys.byteorder)
        for frame in data])
    data = data.astype(float).reshape(-1, channels)

    data -= 128 if bits == 8 else 0
    data = (data + 0.5) / (scaler + 0.5)
    data = data.clip(-1, +1)

    return data.flatten() if channels == 1 else data, sr


def write(path, data, sr, bits=32):
    """
    Writes a .wav file.

    Parameters
    ----------
    path : string
        File path with or without the .wav extension.
    data : ndarray
        Content of the .wav file.
    sr : integer
        Sample rate in hertz.
    bits : integer, optional
        Sample bitwidth.
    """

    if path.startswith('~'):
        path = os.path.expanduser(path)

    if not path.lower().endswith('.wav'):
        path += '.wav'

    data = numpy.asarray(data)
    assert data.dtype in [float, complex]
    assert data.ndim in [1, 2]
    assert data.size > 0

    if numpy.iscomplex(data).any():
        assert data.ndim == 1
        data = numpy.stack((numpy.real(data), numpy.imag(data)))

    if data.ndim == 2:
        if data.shape[0] == 2:
            channels = data.shape[0]
            data = data.ravel('F')
        else:
            channels = data.shape[1]
            data = data.ravel('C')
    else:
        channels = 1

    assert bits in [8, 16, 24, 32]
    bytes = bits // 8
    scaler = 2 ** (bits - 1) - 1

    data = data.clip(-1, +1)
    data = (data * (scaler + 0.5)) - 0.5
    data += 128 if bits == 8 else 0

    data = b''.join([
        int(frame).to_bytes(length=bytes, signed=(bits != 8), byteorder=sys.byteorder)
        for frame in data])

    with wave.open(path, 'wb') as file:
        file.setframerate(sr)
        file.setsampwidth(bytes)
        file.setnchannels(channels)
        file.writeframes(data)



from stftpitchshift.resampler import linear as resample

import numpy as np


def shiftpitch(frames, factors):

    for i in range(len(frames)):

        magnitudes = np.real(frames[i])
        frequencies = np.imag(frames[i])

        magnitudes = np.vstack([resample(magnitudes, factor) for factor in factors])
        frequencies = np.vstack([resample(frequencies, factor) * factor for factor in factors])

        mask = np.argmax(magnitudes, axis=0)

        magnitudes = np.take_along_axis(magnitudes, mask[None,:], axis=0)
        frequencies = np.take_along_axis(frequencies, mask[None,:], axis=0)

        frames[i] = magnitudes + 1j * frequencies

    return frames



import numpy as np


def lowpass(cepstrum, quefrency):

    cepstrum[1:quefrency] *= 2
    cepstrum[quefrency+1:] = 0

    return cepstrum


def lifter(frames, quefrency):

    envelopes = np.zeros(frames.shape)

    for i, frame in enumerate(frames):

        with np.errstate(divide='ignore', invalid='ignore'):
            spectrum = np.log10(np.real(frame))

        cepstrum = np.fft.irfft(spectrum, norm='forward')
        cepstrum = lowpass(cepstrum, quefrency)
        cepstrum = np.fft.rfft(cepstrum, norm='forward')

        envelopes[i] = np.power(10, np.real(cepstrum))

    return envelopes



import numpy as np


def wrap(x):

    return (x + np.pi) % (2 * np.pi) - np.pi


def encode(frames, framesize, hopsize, samplerate):

    M, N = frames.shape

    phaseinc = 2 * np.pi * hopsize / framesize
    freqinc = samplerate / framesize

    buffer = np.zeros(N)
    data = np.zeros((M, N), complex)

    for m, frame in enumerate(frames):

        abs = np.abs(frame)
        arg = np.angle(frame)

        delta = arg - buffer
        buffer = arg

        i = np.arange(N)
        j = wrap(delta - i * phaseinc) / phaseinc

        freq = (i + j) * freqinc

        data[m] = abs + 1j * freq

    return data


def decode(frames, framesize, hopsize, samplerate):

    M, N = frames.shape

    phaseinc = 2 * np.pi * hopsize / framesize
    freqinc = samplerate / framesize

    buffer = np.zeros(N)
    data = np.zeros((M, N), complex)

    for m, frame in enumerate(frames):

        abs = np.real(frame)
        freq = np.imag(frame)

        i = np.arange(N)
        j = (freq - i * freqinc) / freqinc

        delta = (i + j) * phaseinc

        buffer += delta
        arg = buffer

        data[m] = abs * np.exp(1j * arg)

    return data
